!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
module mod_fij
use mod_global, only : dp
implicit none
save
private
public fij,              &
       allocate_fij,     &
       process_dedr_fij, &
       deallocate_fij

real(kind=dp),dimension(:,:),allocatable :: fij

contains

    subroutine allocate_fij
    use mod_neigh, only  : list_length
    use mod_box, only    : dimn
    implicit none

    allocate(fij(dimn,list_length))
    fij = 0.0_dp

    end subroutine allocate_fij

    subroutine process_dedr_fij(dedr,at_l,at_m)
    use mod_neigh, only     : list,marker1,marker2
    use mod_box, only       : dimn,samplesize,pbc
    use mod_particles, only : pos
    implicit none

    ! Passed variables
    real(kind=dp),intent(in) :: dedr
    integer,      intent(in) :: at_l,at_m

    ! Local variables
    integer j,k
    real(kind=dp) :: rij(dimn),distij

    do j = marker1(at_l), marker2(at_l)
        k = list(j)
        if (at_m .eq. k) then
            rij = pos(:,at_l) - pos(:,at_m)
            where(abs(rij) > 0.5_dp * samplesize .and. pbc)
                rij = rij - sign(samplesize,rij)
            end where
            distij = dot_product(rij,rij)
            fij(:,j) = fij(:,j) - dedr * rij / sqrt(distij)
            exit
        endif
    enddo

    return
    end subroutine process_dedr_fij

    subroutine deallocate_fij
    implicit none
    if (allocated(fij)) deallocate(fij)
    end subroutine deallocate_fij

end module mod_fij

