!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
module mod_grid
use mod_global, only : dp
use mod_box, only : dimn
implicit none
save
private
public read_grid,       &
       set_grid,        &
       ngrid,           &
       deallocate_grid, &
       gpt

real(kind=dp), dimension(:,:), allocatable :: gpt

real(kind=dp), dimension(3) :: gpos = 0.0_dp, &
                               gpos_cauchy = 0.0_dp, &
                               gpos_pk1 = 0.0_dp

real(kind=dp)      :: glenx = 0.0_dp, gleny = 0.0_dp, glenz = 0.0_dp, &
                      glenx_cauchy = 0.0_dp, gleny_cauchy = 0.0_dp, glenz_cauchy = 0.0_dp, &
                      glenx_pk1 = 0.0_dp, gleny_pk1 = 0.0_dp, glenz_pk1 = 0.0_dp

integer            :: gdivx = 0, gdivy = 0, gdivz = 0, &
                      gdivx_cauchy = 0, gdivy_cauchy = 0, gdivz_cauchy = 0, &
                      gdivx_pk1 = 0, gdivy_pk1 = 0, gdivz_pk1 = 0, &
                      ngrid = 0

character(len=512) :: gfilename = 'grid_cauchy.data', &
                      gfilename_cauchy = 'grid_cauchy.data', &
                      gfilename_pk1 = 'grid_pk1.data'

logical            :: gfit = .false., gdef = .false., gfile = .false., &
                      gfit_cauchy = .false., gdef_cauchy = .false., gfile_cauchy = .false., &
                      gfit_pk1 = .false., gdef_pk1 = .false., gfile_pk1 = .false.

integer            :: flag_cauchy = 0, flag_pk1 = 0
! Tips: we can use gdef x,0,0 or 0,y,0 or 0,0,z to define 1D grid

contains

    subroutine read_grid
    use mod_box, only : dimn
    use mod_output, only : string,output,erroroutput,nerror,error,heading
    use mod_global, only : cmdunit
    use mod_qclib, only : next,downcase,freein
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=512) :: input
    character(len=5) :: cc
    character(len=4) :: subcc
    integer :: lower,upper,i,ldif,idum,ierror
    real(kind=dp) :: rdum
    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    heading=.true.
    write(string,'(a)') '**Reading grid data**'; call output
    do
        read(cmdunit,'(a512)',iostat=ierror) input
        if (ierror .ne. 0) then
            nerror = 2  
            write(error(1),'(a)') 'Something wrong with the command input.'
            write(error(2),'(a)') 'Macro end must be called to continue the simulation.'
            call erroroutput
        end if
        lower = 0
        upper = next(lower,input)
        cc = input(lower+1:upper-1)
        select case(trim(downcase(cc)))
        case('gfit')
            ! Set gdivx,gdivy,gdivz
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivx,rdum,1)
            if (gdivx < 0) then
                nerror = 2
                write(error(1),'(a)') 'gdivx must be >= 0.'
                write(error(2),'(a,I0)') 'Your input: ', gdivx
                call erroroutput
            end if
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivy,rdum,1)
            if (gdivy < 0) then
                nerror = 2
                write(error(1),'(a)') 'gdivy must be >= 0.'
                write(error(2),'(a,I0)') 'Your input: ', gdivy
                call erroroutput
            end if
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivz,rdum,1)
            if (gdivz < 0) then
                nerror = 2
                write(error(1),'(a)') 'gdivz must be >= 0.'
                write(error(2),'(a,I0)') 'Your input: ', gdivz
                call erroroutput
            end if

            lower = upper
            upper = next(lower,input)
            subcc = input(lower+1:upper-1)

            select case(trim(downcase(subcc)))
                case('cauc')
                    if (flag_cauchy == 0) then
                        flag_cauchy = 1
                    else
                        gdef_cauchy = .false.
                        gfile_cauchy = .false.
                        write(string,'(a)') 'Warning: Grid point definition for Cauchy stress has already been given!'
                        call output
                        write(string,'(a)') 'Resetting grid point for Cauchy stress.'
                        call output
                    end if
                    gfit_cauchy = .true.
                    gdivx_cauchy = gdivx
                    gdivy_cauchy = gdivy
                    gdivz_cauchy = gdivz
                    lower = upper
                    upper = next(lower,input)
                    ldif = upper - lower - 1
                    if (ldif /= 0) then
                        gfilename_cauchy = trim(input(lower+1:upper-1)) // '.data'
                    endif
                    write(string,3000) gfit_cauchy; call output
                    write(string,3500) gdivx_cauchy; call output
                    write(string,3600) gdivy_cauchy; call output
                    write(string,3700) gdivz_cauchy; call output
                    write(string,3009) trim(gfilename_cauchy); call output
                case('pk1')
                    if (flag_pk1 == 0) then
                        flag_pk1 = 1
                    else
                        gdef_pk1 = .false.
                        gfile_pk1 = .false.
                        write(string,'(a)') 'Warning: Grid point definition for 1st pk stress has already been given!'
                        call output
                        write(string,'(a)') 'Resetting grid point for 1st pk stress.'
                        call output
                    end if
                    gfit_pk1 = .true.
                    gdivx_pk1 = gdivx
                    gdivy_pk1 = gdivy
                    gdivz_pk1 = gdivz
                    lower = upper
                    upper = next(lower,input)
                    ldif = upper - lower - 1
                    if (ldif /= 0) then
                        gfilename_pk1 = trim(input(lower+1:upper-1)) // '.data'
                    endif
                    write(string,3001) gfit_pk1; call output
                    write(string,3501) gdivx_pk1; call output
                    write(string,3601) gdivy_pk1; call output
                    write(string,3701) gdivz_pk1; call output
                    write(string,3010) trim(gfilename_pk1); call output
                case('')
                    if (flag_cauchy == 0) then
                        flag_cauchy = 1
                    else
                        gdef_cauchy = .false.
                        gfile_cauchy = .false.
                        write(string,'(a)') &
                        'Warning: Grid point definition for Cauchy stress has already been given!'
                        call output
                        write(string,'(a)') 'Resetting grid point for Cauchy stress.'
                        call output
                    end if

                    if (flag_pk1 == 0) then
                        flag_pk1 = 1
                    else
                        gdef_pk1 = .false.
                        gfile_pk1 = .false.
                        write(string,'(a)') &
                        'Warning: Grid point definition for 1st pk stress has already been given!'
                        call output
                        write(string,'(a)') 'Resetting grid point for 1st pk stress.'
                        call output
                    end if

                    gfit = .true.
                    gfit_cauchy = .true.
                    gdivx_cauchy = gdivx
                    gdivy_cauchy = gdivy
                    gdivz_cauchy = gdivz
                    gfit_pk1 = .true.
                    gdivx_pk1 = gdivx
                    gdivy_pk1 = gdivy
                    gdivz_pk1 = gdivz

                    lower = upper
                    upper = next(lower,input)
                    ldif = upper - lower - 1
                    if (ldif /= 0) then
                        gfilename_cauchy = trim(input(lower+1:upper-1)) // '.data'
                    endif

                    lower = upper
                    upper = next(lower,input)
                    ldif = upper - lower - 1
                    if (ldif /= 0) then
                        gfilename_pk1 = trim(input(lower+1:upper-1)) // '.data'
                    endif

                    write(string,3002) gfit; call output
                    write(string,3502) gdivx; call output
                    write(string,3602) gdivy; call output
                    write(string,3702) gdivz; call output
                    write(string,3009) trim(gfilename_cauchy); call output
                    write(string,3010) trim(gfilename_pk1); call output 

                case default
                    nerror = 1
                    error(1) = 'sub macro "' // subcc // '" not recognized in macro gfit**'
                    call erroroutput
            end select
        case('gdef')
            ! Read gpos
            do i = 1, dimn
                lower=upper
                upper=next(lower,input)
                call freein(input,lower,upper,idum,gpos(i),2)
            end do
            ! Set glenx,gleny,glenz
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,idum,glenx,2)
            if (glenx < 0.0_dp .or. glenx > 1.0_dp) then
                nerror = 2
                write(error(1),'(a)') 'glenx must lie in [0,1].'
                write(error(2),'(a,F25.16)') 'Your input: ', glenx
                call erroroutput
            end if
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,idum,gleny,2)
            if (gleny < 0.0_dp .or. gleny > 1.0_dp) then
                nerror = 2
                write(error(1),'(a)') 'gleny must lie in [0,1].'
                write(error(2),'(a,F25.16)') 'Your input: ', gleny
                call erroroutput
            end if
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,idum,glenz,2)
            if (glenz < 0.0_dp .or. glenz > 1.0_dp) then
                nerror = 2
                write(error(1),'(a)') 'glenz must lie in [0,1].'
                write(error(2),'(a,F25.16)') 'Your input: ', glenz
                call erroroutput
            end if
            ! Set gdivx,gdivy,gdivz
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivx,rdum,1)
            if (gdivx < 0) then
                nerror = 2
                write(error(1),'(a)') 'gdivx must be >= 0.'
                write(error(2),'(a,I0)') 'Your input: ', gdivx
                call erroroutput
            end if
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivy,rdum,1)
            if (gdivy < 0) then
                nerror = 2
                write(error(1),'(a)') 'gdivy must be >= 0.'
                write(error(2),'(a,I0)') 'Your input: ', gdivy
                call erroroutput
            end if
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivz,rdum,1)
            if (gdivz < 0) then
                nerror = 2
                write(error(1),'(a)') 'gdivz must be >= 0.'
                write(error(2),'(a,I0)') 'Your input: ', gdivz
                call erroroutput
            end if

            lower = upper
            upper = next(lower,input)
            subcc = input(lower+1:upper-1)

            select case(trim(downcase(subcc)))
                case('cauc')
                    if (flag_cauchy == 0) then
                        flag_cauchy = 1
                    else
                        gfit_cauchy = .false.
                        gfile_cauchy = .false.
                        write(string,'(a)') 'Warning: Grid point definition for Cauchy stress has already been given!'
                        call output
                        write(string,'(a)') 'Resetting grid point for Cauchy stress.'
                        call output
                    end if

                    gdef_cauchy = .true.
                    gpos_cauchy = gpos
                    glenx_cauchy = glenx
                    gleny_cauchy = gleny
                    glenz_cauchy = glenz
                    gdivx_cauchy = gdivx
                    gdivy_cauchy = gdivy
                    gdivz_cauchy = gdivz
                    lower = upper
                    upper = next(lower,input)
                    ldif = upper - lower - 1
                    if (ldif /= 0) then
                        gfilename_cauchy = trim(input(lower+1:upper-1)) // '.data'
                    endif

                    write(string,3003)  gdef_cauchy; call output
                    write(string,3100)  gpos_cauchy; call output
                    write(string,3200) glenx_cauchy; call output
                    write(string,3300) gleny_cauchy; call output
                    write(string,3400) glenz_cauchy; call output
                    write(string,3500) gdivx_cauchy; call output
                    write(string,3600) gdivy_cauchy; call output
                    write(string,3700) gdivz_cauchy; call output
                    write(string,3009) trim(gfilename_cauchy); call output
                case('pk1')
                    if (flag_pk1 == 0) then
                        flag_pk1 = 1
                    else
                        gfit_pk1 = .false.
                        gfile_pk1 = .false.
                        write(string,'(a)') 'Warning: Grid point definition for 1st pk stress has already been given!'
                        call output
                        write(string,'(a)') 'Resetting grid point for 1st pk stress.'
                        call output
                    end if
                    gdef_pk1 = .true.
                    gpos_pk1 = gpos
                    glenx_pk1 = glenx
                    gleny_pk1 = gleny
                    glenz_pk1 = glenz
                    gdivx_pk1 = gdivx
                    gdivy_pk1 = gdivy
                    gdivz_pk1 = gdivz
                    lower = upper
                    upper = next(lower,input)
                    ldif = upper - lower - 1
                    if (ldif /= 0) then
                        gfilename_pk1 = trim(input(lower+1:upper-1)) // '.data'
                    endif

                    write(string,3004)  gdef_pk1; call output
                    write(string,3101)  gpos_pk1; call output
                    write(string,3201) glenx_pk1; call output
                    write(string,3301) gleny_pk1; call output
                    write(string,3401) glenz_pk1; call output
                    write(string,3501) gdivx_pk1; call output
                    write(string,3601) gdivy_pk1; call output
                    write(string,3701) gdivz_pk1; call output
                    write(string,3010) trim(gfilename_pk1); call output

                case('')
                    if (flag_cauchy == 0) then
                        flag_cauchy = 1
                    else
                        gfit_cauchy = .false.
                        gfile_cauchy = .false.
                        write(string,'(a)') &
                        'Warning: Grid point definition for Cauchy stress has already been given!'
                        call output
                        write(string,'(a)') 'Resetting grid point for Cauchy stress.'
                        call output
                    end if

                    if (flag_pk1 == 0) then
                        flag_pk1 = 1
                    else
                        gfit_pk1 = .false.
                        gfile_pk1 = .false.
                        write(string,'(a)') &
                        'Warning: Grid point definition for 1st pk stress has already been given!'
                        call output
                        write(string,'(a)') 'Resetting grid point for 1st pk stress.'
                        call output
                    end if

                    gdef = .true.
                    gdef_cauchy = .true.
                    gpos_cauchy = gpos
                    glenx_cauchy = glenx
                    gleny_cauchy = gleny
                    glenz_cauchy = glenz
                    gdivx_cauchy = gdivx
                    gdivy_cauchy = gdivy
                    gdivz_cauchy = gdivz
                    gdef_pk1 = .true.
                    gpos_pk1 = gpos
                    glenx_pk1 = glenx
                    gleny_pk1 = gleny
                    glenz_pk1 = glenz
                    gdivx_pk1 = gdivx
                    gdivy_pk1 = gdivy
                    gdivz_pk1 = gdivz

                    lower = upper
                    upper = next(lower,input)
                    ldif = upper - lower - 1
                    if (ldif /= 0) then
                        gfilename_cauchy = trim(input(lower+1:upper-1)) // '.data'
                    endif

                    lower = upper
                    upper = next(lower,input)
                    ldif = upper - lower - 1
                    if (ldif /= 0) then
                        gfilename_pk1 = trim(input(lower+1:upper-1)) // '.data'
                    endif

                    write(string,3005) gdef; call output
                    write(string,3102) gpos; call output
                    write(string,3202) glenx; call output
                    write(string,3302) gleny; call output
                    write(string,3402) glenz; call output
                    write(string,3502) gdivx; call output
                    write(string,3602) gdivy; call output
                    write(string,3702) gdivz; call output
                    write(string,3009) trim(gfilename_cauchy); call output
                    write(string,3010) trim(gfilename_pk1); call output 

                case default
                    nerror = 1
                    error(1) = 'sub macro "' // subcc // '" not recognized in macro gdef**'
                    call erroroutput
            end select
        case('gfile')
            lower=upper
            upper=next(lower,input)
            ldif = upper - lower - 1
            if (ldif /= 0) then
                if (flag_cauchy == 0) then
                    flag_cauchy = 1
                else
                    gfit_cauchy = .false.
                    gdef_cauchy = .false.
                    write(string,'(a)') 'Warning: Grid point definition for Cauchy stress has already been given!'
                    call output
                    write(string,'(a)') 'Resetting grid point for Cauchy stress.'
                    call output
                end if
                gfile_cauchy = .true.
                write(string,3006) gfile_cauchy; call output
                gfilename_cauchy = trim(input(lower+1:upper-1)) // '.data'
                write(string,3009) trim(gfilename_cauchy); call output
            endif
            lower = upper
            upper = next(lower,input)
            ldif = upper - lower - 1
            if (ldif /= 0) then
                if (flag_pk1 == 0) then
                    flag_pk1 = 1
                else
                    gfit_pk1 = .false.
                    gdef_pk1 = .false.
                    write(string,'(a)') 'Warning: Grid point definition for 1st pk stress has already been given!'
                    call output
                    write(string,'(a)') 'Resetting grid point for 1st pk stress.'
                    call output
                end if
                gfile_pk1 = .true.
                write(string,3007) gfile_pk1; call output
                gfilename_pk1 = trim(input(lower+1:upper-1)) // '.data'
                write(string,3010) trim(gfilename_pk1); call output
            endif
        case('end')
            if (flag_cauchy == 1 .or. flag_pk1 == 1) then
                exit
            else
                nerror = 1
                write(error(1),'(a)') "Grid option (gfit/gdef/gfile) must be specified in macro grid."
                call erroroutput
            endif
        case default
            nerror = 1
            error(1) = 'macro "' // cc // '" not recognized in stage grid**'
            call erroroutput
        end select
    enddo

    3000 format('Fit the grid Cauchy                      = ',l5)
    3001 format('Fit the grid 1st pk                      = ',l5)
    3002 format('Fit the grid Cauchy and 1st pk           = ',l5)

    3003 format('Define the grid Cauchy                   = ',l5)
    3004 format('Define the grid 1st pk                   = ',l5)
    3005 format('Define the grid Cauchy and 1st pk        = ',l5)

    3006 format('Read grid from file Cauchy               = ',l5)
    3007 format('Read grid from file 1st pk               = ',l5)
    3008 format('Read grid from file Cauchy and 1st pk    = ',l5)

    3009 format('Write to grid file Cauchy                = ',a)
    3010 format('Write to grid file 1st pk                = ',a)

    3100 format('Position of the grid Cauchy              = ',3f25.16)
    3200 format('Length of grid : x-dir Cauchy            = ',f25.16)
    3300 format('                 y-dir Cauchy            = ',f25.16)
    3400 format('                 z-dir Cauchy            = ',f25.16)
    3500 format('Number of divs : x-dir Cauchy            = ',I0)
    3600 format('                 y-dir Cauchy            = ',I0)
    3700 format('                 z-dir Cauchy            = ',I0)

    3101 format('Position of the grid 1st pk              = ',3f25.16)
    3201 format('Length of grid : x-dir 1st pk            = ',f25.16)
    3301 format('                 y-dir 1st pk            = ',f25.16)
    3401 format('                 z-dir 1st pk            = ',f25.16)
    3501 format('Number of divs : x-dir 1st pk            = ',I0)
    3601 format('                 y-dir 1st pk            = ',I0)
    3701 format('                 z-dir 1st pk            = ',I0)

    3102 format('Position of the grid Cauchy and 1st pk   = ',3f25.16)
    3202 format('Length of grid : x-dir Cauchy and 1st pk = ',f25.16)
    3302 format('                 y-dir Cauchy and 1st pk = ',f25.16)
    3402 format('                 z-dir Cauchy and 1st pk = ',f25.16)
    3502 format('Number of divs : x-dir Cauchy and 1st pk = ',I0)
    3602 format('                 y-dir Cauchy and 1st pk = ',I0)
    3702 format('                 z-dir Cauchy and 1st pk = ',I0)

    end subroutine read_grid

    subroutine set_grid(characteristic_length,stress)
    use mod_box, only : dimn,nat,samplesize,pbc
    use mod_particles, only : pos
    use mod_output, only : string,output,heading,erroroutput,nerror,error
    use mod_file, only : FileExists,iofile,GetFreeUnit
    implicit none

    ! Passed variables
    ! characteristic_length is used in gfit to determine the grid fit area
    real(kind=dp), dimension(3), intent(in) :: characteristic_length
    character(len=6), intent(in) :: stress
    ! Local variables
    real(kind=dp), dimension(3) :: gpos_min,gpos_max
    real(kind=dp) :: delx,dely,delz
    integer :: i,j,k,l,gunit,maxlines,ios
    character(len=1) :: junk

    if (trim(stress) == 'cauchy') then
        if (flag_cauchy /= 1) then
            nerror = 2
            write(error(1),'(a)') 'Grid option for cauchy stress is not given!'
            write(error(2),'(a)') 'Program aborted!'
            call erroroutput
        end if
        heading = .true.
        write(string,'(a)')'**Setting the Cauchy stress grid**'; call output
        gfit = gfit_cauchy
        gdef = gdef_cauchy
        gfile = gfile_cauchy
        gpos = gpos_cauchy
        glenx = glenx_cauchy
        gleny = gleny_cauchy
        glenz = glenz_cauchy
        gdivx = gdivx_cauchy
        gdivy = gdivy_cauchy
        gdivz = gdivz_cauchy
        gfilename = gfilename_cauchy
    else if (trim(stress) == 'pk1') then
        if (flag_pk1 /= 1) then
            nerror = 2
            write(error(1),'(a)') 'Grid option for 1st pk stress is not given!'
            write(error(2),'(a)') 'Program aborted!'
            call erroroutput
        end if
        heading = .true.
        write(string,'(a)')'**Setting the 1st pk stress grid**'; call output
        gfit = gfit_pk1
        gdef = gdef_pk1
        gfile = gfile_pk1
        gpos = gpos_pk1
        glenx = glenx_pk1
        gleny = gleny_pk1
        glenz = glenz_pk1
        gdivx = gdivx_pk1
        gdivy = gdivy_pk1
        gdivz = gdivz_pk1
        gfilename = gfilename_pk1
    else
        nerror = 1
        write(error(1),'(a)') 'This code now could only calculate Cauchy and 1st pk stress.'
        write(error(2),'(a)') 'Error creating the grid point. Program Aborted!'
        call erroroutput
    end if

    if (gfit .or. gdef) then
        gunit = GetFreeUnit()
        open(gunit,file=gfilename,status='replace',form='formatted')
        write(string,'(2a)') 'Writing grid points to file ',&
                              trim(gfilename); call output
        delx = 0.0_dp
        dely = 0.0_dp
        delz = 0.0_dp

        if (gfit) then
            gpos_min = huge(1.0_dp)
            gpos_max = -huge(1.0_dp)

            do i = 1, nat
                where(pos(:,i) .lt. gpos_min)
                    gpos_min = pos(:,i)
                end where
                where(pos(:,i) .gt. gpos_max)
                    gpos_max = pos(:,i)
                end where
            end do

            gpos_min = gpos_min + characteristic_length*(/(1.0_dp,i=1,dimn)/)
            gpos_max = gpos_max - characteristic_length*(/(1.0_dp,i=1,dimn)/)

            do i = 1, dimn
                !if (gpos_min(i) .gt. gpos_max(i) .and. pbc(i)) then
                if (gpos_min(i) .gt. gpos_max(i)) then
                    nerror=3
                    write(error(1),'(a)') &
                    'Error in creating the grid point.'
                    write(error(2),'(a)') &
                    'Either decrease the averaging domain size or increase the sample size.'
                    write(error(3),'(a,I0,a,F25.16,a,F25.16)') &
                        'In ', i, ' direction: ', gpos_min(i), 'is smaller than ', gpos_max(i)
                    call erroroutput
                 end if
            end do

            gpos = gpos_min

            if (gdivx .eq. 0) gpos(1) = 0.5_dp*(gpos_min(1)+gpos_max(1))
            if (gdivy .eq. 0) gpos(2) = 0.5_dp*(gpos_min(2)+gpos_max(2))
            if (gdivz .eq. 0) gpos(3) = 0.5_dp*(gpos_min(3)+gpos_max(3))


            if (gdivx .ne. 0) delx = (gpos_max(1)-gpos_min(1))/gdivx
            if (gdivy .ne. 0) dely = (gpos_max(2)-gpos_min(2))/gdivy
            if (gdivz .ne. 0) delz = (gpos_max(3)-gpos_min(3))/gdivz

        end if

        if (gdef) then
            if (glenx .eq. 0._dp) gdivx = 0
            if (gleny .eq. 0._dp) gdivy = 0
            if (glenz .eq. 0._dp) gdivz = 0

            if (gdivx .ne. 0) delx = glenx*samplesize(1)/gdivx
            if (gdivy .ne. 0) dely = gleny*samplesize(2)/gdivy
            if (gdivz .ne. 0) delz = glenz*samplesize(3)/gdivz
        end if

        ngrid = (gdivx+1)*(gdivy+1)*(gdivz+1)
        write(string,3080) ngrid; call output

        if (allocated(gpt)) deallocate(gpt)
        if (.not. allocated(gpt)) allocate(gpt(dimn,ngrid))

        l = 1
        do i = 1,gdivx + 1
            do j = 1,gdivy + 1
                do k = 1,gdivz + 1
                    gpt(:,l) = gpos + (/(i-1)*delx,  &
                                        (j-1)*dely,  &
                                        (k-1)*delz/)
                    write(gunit,'(3F25.16)') gpt(:,l)
                    l = l + 1
                enddo
            enddo
        enddo
        close(gunit)
    endif

    if (gfile) then
        if (FileExists(gfilename,.false.)) then
            call iofile(gfilename,'formatted  ',gunit,.false.,.false.)
            heading=.true.
            write(string,'(a)') '**Reading grid data from file**'
            call output
        else   
            nerror = 2
            error(1) = 'grid data file not found'
            error(2) = 'in subroutine set_grid, filename = ' &
                        // trim(gfilename)
            call erroroutput
        endif
        
        write(string,'(2a)') 'Reading grid points from file ',&
                              trim(gfilename); call output
 
        ngrid = 0
        maxlines = 100000000
        do i = 1,maxlines
            read(gunit,*,iostat=ios) junk
            if (ios .ne. 0) then
                exit
            endif

            if (i .eq. maxlines) then
                nerror = 2
                error(1) = 'Maximum lines (default = 1000000000) reached.'
                error(2) = 'If you do have so many lines, modify maxlines in mod_grid.f90'
                call erroroutput
            endif
            ngrid  = ngrid + 1
        enddo

        if (ngrid == 0) then
            nerror = 1
            write(error(1),'(a)') 'Empty grid data file! Simulation aborted!'
            call erroroutput
        end if

        write(string,3080) ngrid; call output

        rewind(gunit)

        if (allocated(gpt)) deallocate(gpt)
        if (.not. allocated(gpt)) allocate(gpt(dimn,ngrid))

        do i = 1,ngrid
            read(gunit,*) gpt(:,i)
        enddo

        close(gunit)
    endif

    ! Do we allow grid point outside the model?
    ! like the grid point outside the boundary of the model?
    !
    if (gdef .or. gfile) then
        call compatibility_check(characteristic_length)
    end if

    write(string,'(a)') 'Grid points set finished!'; call output

    3080 format('Number of grid points  = ',I10)
    end subroutine set_grid

    ! This could be done in parallel
    subroutine compatibility_check(characteristic_length)
    use mod_box, only : dimn,nat,pbc
    use mod_particles, only : pos
    use mod_output, only : erroroutput,nerror,error
    implicit none

    ! Passed variables
    ! characteristic_length is used in gfit to determine the grid fit area
    real(kind=dp), dimension(3), intent(in) :: characteristic_length

    ! Local variables
    real(kind=dp), dimension(dimn) :: pos_max, pos_min
    integer :: i,l

    pos_max = -huge(1.0_dp)
    pos_min = huge(1.0_dp)

    do i = 1, nat
        where (pos(:,i) > pos_max)
            pos_max = pos(:,i)
        end where
        where (pos(:,i) < pos_min)
            pos_min = pos(:,i)
        end where
    end do

    do l = 1, ngrid
        do i = 1, dimn
            if (gpt(i,l) > pos_max(i) - characteristic_length(i) .and. pbc(i)) then
                if (i == 1) then
                    nerror = 1
                    write(error(1),'(a,I0,a,F25.16,a,F25.16)') &
                        'In x direction which is periodic, grid point ', l, " : ",&
                        gpt(i,l), ' is bigger than the maximum atomic coordinate - characteristic_length(1) ',&
                        pos_max(i) - characteristic_length(i)
                    call erroroutput
                else if (i == 2) then
                    nerror = 1
                    write(error(1),'(a,I0,a,F25.16,a,F25.16)') &
                        'In y direction which is periodic, grid point ', l, " : ",&
                        gpt(i,l), ' is bigger than the maximum atomic coordinate - characteristic_length(2) ',&
                        pos_max(i) - characteristic_length(i)
                    call erroroutput
                else if (i == 3) then
                    nerror = 1
                    write(error(1),'(a,I0,a,F25.16,a,F25.16)') &
                        'In z direction which is periodic, grid point ', l, " : ",&
                        gpt(i,l), ' is bigger than the maximum atomic coordinate - characteristic_length(3) ',&
                        pos_max(i) - characteristic_length(i)
                    call erroroutput
                else
                    nerror = 1
                    write(error(1),'(a)') 'Something funny happens. dimn /= 3.'
                    call erroroutput
                end if
            end if

            if (gpt(i,l) < pos_min(i) + characteristic_length(i) .and. pbc(i)) then
                if (i == 1) then
                    nerror = 1
                    write(error(1),'(a,I0,a,F25.16,a,F25.16)') &
                        'In x direction which is periodic, grid point ', l, " : ",&
                        gpt(i,l), ' is smaller than the minimum atomic coordinate + characteristic_length(1) ',&
                        pos_min(i) + characteristic_length(i)
                    call erroroutput
                else if (i == 2) then
                    nerror = 1
                    write(error(1),'(a,I0,a,F25.16,a,F25.16)') &
                        'In y direction which is periodic, grid point ', l, " : ",&
                        gpt(i,l), ' is smaller than the minimum atomic coordinate + characteristic_length(2) ',&
                        pos_min(i) + characteristic_length(i)
                    call erroroutput
                else if (i == 3) then
                    nerror = 1
                    write(error(1),'(a,I0,a,F25.16,a,F25.16)') &
                        'In z direction which is periodic, grid point ', l, " : ",&
                        gpt(i,l), ' is smaller than the minimum atomic coordinate + characteristic_length(3) ',&
                        pos_min(i) + characteristic_length(i)
                    call erroroutput
                else
                    nerror = 1
                    write(error(1),'(a)') 'Something funny happens. dimn /= 3.'
                    call erroroutput
                end if
            end if
        end do
    end do

    end subroutine compatibility_check

    subroutine deallocate_grid
    implicit none
    if (allocated(gpt)) deallocate(gpt)
    end subroutine deallocate_grid

end module mod_grid

