!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!-------------------------------------------------------------------------------
! MODULE :
!
! type definitions:
! ----------------
! N/A
!
! Real Variable Definitions:
! -------------------------------------
!
! Integer Variable Definitions:
! ----------------------------
!
! logical Variable Definitions:
! ----------------------------
!
! Contains Routines:
! ------------------
!
! SUBROUTINE :
!
! Passed Parameters :-
! -----------------
!
! Non-Obvious Parameter :-
! ---------------------
!
module mod_hardy
use mod_global, only : dp, qp, PI
implicit none
save
private
public hardy,               &
       hardy_i,             & ! projection -> Irrotational part
       hardy_kin,           & ! kinetic
       avgdomain_hardy,     &
       collect_hardy,       &
       write_hardy_cauchy,  &
       write_hardy_pk1,     &
       hardy_radius,        &
       hardy_volume,        &
       virial,              &
       virial_i,            &
       virial_kin,          &
       avgdomain_virial,    &
       write_virial_cauchy, &
       write_virial_pk1,    &
       virial_radius,       &
       virial_volume

real(kind=dp), dimension(:,:,:), allocatable :: hardy, hardy_i, hardy_kin, &
                                                virial, virial_i, virial_kin

character(len=512)              :: avgdomain_hardy = 'sphere', avgdomain_virial = 'sphere'
! For spherical average domain
real(kind=dp)                   :: hardy_radius = 1.0_dp, virial_radius = 1.0_dp 

real(kind=dp)                   :: hardy_volume = 4.0_dp * PI / 3.0_dp, virial_volume = 4.0_dp * PI / 3.0_dp

contains

   subroutine collect_hardy(l,crossings,cutoff,fij,list,marker1,marker2,tensor,tensork,avgdomain,wfunc)
   use mod_box, only : dimn,samplesize,pbc!,nat
   use mod_grid, only : gpt
   use mod_particles, only : pos,vel
   use mod_neigh, only : full_grid_verlet_list
   use mod_qclib, only : tensor_prod
   use mod_output, only : erroroutput, nerror, error
   use lsqrblasInterface, only : ddot
   implicit none
   ! Passed variables
   integer,       intent(in)      :: l
   logical,       intent(in)      :: crossings
   integer,       intent(in)      :: marker1(:),marker2(:),list(:)
   real(kind=dp), intent(in)      :: fij(:,:),cutoff
   character(len=512)             :: avgdomain, wfunc
   real(kind=dp), intent(out)     :: tensor(:,:),tensork(:,:)
   ! Local variables
   real(kind=dp), dimension(dimn) :: r,rij,p1,p2,rp1,rp2,ra,rb,avgvel,rel_vel
   real(kind=dp), dimension(dimn,3) :: ri,rj
   real(kind=dp), dimension(dimn,dimn) :: rcrossr,vcrossv
   real(kind=dp) :: eff_bond = 0.0_dp,eff_bond_total = 0.0_dp
   real(kind=dp) :: dist,dist_ri,dist_rj,dist_rij,dist_r,lambda,rsqsp,s,eff_line,rsumsq
   real(kind=dp) :: eps = 1.0e-5_dp
   integer :: i,j,k,m,n,p,interior_cnt,interior(full_grid_verlet_list(l)%nneigh)
   integer :: debug = 0

   avgdomain = trim(adjustl(avgdomain))
   wfunc = trim(adjustl(wfunc))

   tensor = 0.0_dp
   tensork = 0.0_dp
   eff_bond_total = 0.0_dp

   dist_rij = 0.0_dp ! Just to suppress warnings

   if (crossings) then
       rsqsp = hardy_radius**2
       rsumsq = (hardy_radius + cutoff)**2
   else
       rsqsp = virial_radius**2
       rsumsq = (virial_radius + cutoff)**2
   end if

   interior = 0
   interior_cnt = 0
   avgvel = 0.0_dp

   do n = 1, full_grid_verlet_list(l)%nneigh

       i = full_grid_verlet_list(l)%neigh(n)
       r = pos(:,i)-gpt(:,l)

       where (abs(r)/samplesize > 0.5_dp .and. pbc)
           r = r-sign(samplesize,r)
       end where
       dist = ddot(dimn,r,1,r,1)

       ! Collect the avg domain's interior particles to account for kinetic contribution
       if (dist .le. rsqsp) then
           interior_cnt = interior_cnt + 1
           interior(interior_cnt) = i
           avgvel=avgvel+vel(:,i)
       end if
       ! End of kinetic contribution

       if (dist .le. rsumsq) then
           do p = marker1(i), marker2(i)
           !-------------------------------------------------------------------------------
           !
           !                 ==============================
           !                ||                            ||
           !                ||  _______                   ||
           !                || / .p1   \                  ||  .rp1
           !                ||/ /       \                 || /
           !                ||\/  Avg.  /                 ||/
           !                ||/\ domain/                  |/
           !                |/  \     /                   /|
           !                /|   \___/                   /||
           !               /||                          / ||
           !           rp2. ||                       p2.  ||
           !                ||                            ||
           !                 ==============================
           !
           !
               j = list(p)
               p1 = pos(:,i)
               p2 = pos(:,j)
               rij = p1-p2
               where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                   rij = rij-sign(samplesize,rij)
               end where
               rp1 = p2 + rij
               rp2 = p1 - rij
               if (maxval(abs(rp1-p1)) .lt. eps) then
                   k=1
               else
                   k=2
               endif

               ri(:,1) = p1  - gpt(:,l)
               rj(:,1) = rp2 - gpt(:,l)
               ri(:,2) = rp1 - gpt(:,l)
               rj(:,2) = p2  - gpt(:,l)

               eff_bond = 0.0_dp

               do m = 1, k

                   dist_ri = ddot(dimn,ri(:,m),1,ri(:,m),1)
                   dist_rj = ddot(dimn,rj(:,m),1,rj(:,m),1)
                   if (dist_ri .gt. rsumsq .or. dist_rj .gt. rsumsq) then
                       cycle
                   endif
                   dist_rij = ddot(dimn,rij,1,rij,1)
                   if (dist_rij > epsilon(1.0_dp)) then
                       lambda = ddot(dimn,ri(:,m),1,rij,1) / dist_rij
                   else
                       nerror = 2
                       write(error(1),'(a)') "dist_rij == 0 indicating you might have overlapping atoms in your system!"
                       write(error(2),'(a,I0,a,I0,a,3F10.5)') "The two atoms ", i, " and ", j, &
                                                              " coincide at: ", pos(:,i)
                       call erroroutput
                   endif
                   r = ri(:,m) - lambda * rij
                   dist_r = ddot(dimn,r,1,r,1)

                   if (abs(rsqsp - dist_r) .lt. eps) then
                       cycle
                   endif
                   !-------------------------------------------------------------------------------
                   ! If both particles lie within the Hardy sphere
                   !
                   if ((dist_ri .le. rsqsp) .and. (dist_rj .le. rsqsp)) then
                       ra = ri(:,m)
                       rb = rj(:,m)
                       if (crossings .eqv. .false.) then
                           eff_bond = sqrt(dist_rij) / virial_volume
                           eff_bond_total = eff_bond_total + eff_bond
                           cycle
                       else
                           if (wfunc == 'constant') then
                               eff_bond = sqrt(dist_rij) / hardy_volume
                               eff_bond_total = eff_bond_total + eff_bond
                               cycle
                           else if (wfunc == 'mollify') then
                               call qtrap(ra,rb,eff_bond,debug)
                               eff_bond_total = eff_bond_total + eff_bond
                               cycle
                           else
                               nerror = 2
                               write(error(1),'(2a)') wfunc, ' weighting function is not supported in the code!'
                               write(error(2),'(a)') 'Program aborted!'
                               call erroroutput
                           end if
                       endif
                   endif
                   if (crossings .eqv. .true.) then
                       !-------------------------------------------------------------------------------
                       ! If both particles lie outside the Hardy sphere
                       !
                       if((dist_ri .gt. rsqsp) .and. (dist_rj .gt. rsqsp) .and. &
                          (lambda .gt. 0.0_dp) .and. (lambda .lt. 1.0_dp) .and. &
                          (dist_r .lt. rsqsp)) then

                           eff_line = 2._dp * sqrt(rsqsp - dist_r)
                           ra = r + 0.5_dp * eff_line * (rj(:,m) - ri(:,m))/sqrt(dist_rij)
                           rb = r - 0.5_dp * eff_line * (rj(:,m) - ri(:,m))/sqrt(dist_rij)
                           if (wfunc == 'constant') then
                               eff_bond = eff_bond + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                               eff_bond_total = eff_bond_total + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                           else if (wfunc == 'mollify') then
                               call qtrap(ra,rb,s,debug)
                               eff_bond = eff_bond + s
                               eff_bond_total = eff_bond_total + s
                           else
                               nerror = 2
                               write(error(1),'(2a)') wfunc, ' weighting function is not supported in the code!'
                               write(error(2),'(a)') 'Program aborted!'
                               call erroroutput
                           end if
                       endif
                       !-------------------------------------------------------------------------------
                       ! If one particle lies inside and other lies outside the Hardy sphere
                       !
                       if ((dist_ri .le. rsqsp) .and. (dist_rj .gt. rsqsp)) then
                           ra = ri(:,m)
                           if (dist_ri - dist_r .gt. eps) then
                               eff_line = sqrt(rsqsp - dist_r) +                   &
                                          sign(sqrt(dist_ri - dist_r),lambda)
                               rb = ri(:,m) + eff_line*(rj(:,m) - ri(:,m))/sqrt(dist_rij)
                               if (wfunc == 'constant') then
                                   eff_bond = eff_bond + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                                   eff_bond_total = eff_bond_total + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                               else if (wfunc == 'mollify') then
                                   call qtrap(ra,rb,s,debug)
                                   eff_bond = eff_bond + s
                                   eff_bond_total = eff_bond_total + s
                               else
                                   nerror = 2
                                   write(error(1),'(2a)') wfunc, ' weighting function is not supported in the code!'
                                   write(error(2),'(a)') 'Program aborted!'
                                   call erroroutput
                               end if
                           else
                               eff_line = sqrt(rsqsp - dist_r)
                               rb = ri(:,m) + eff_line*(rj(:,m) - ri(:,m))/sqrt(dist_rij)
                               if (wfunc == 'constant') then
                                   eff_bond = eff_bond + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                                   eff_bond_total = eff_bond_total + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                               else if (wfunc == 'mollify') then
                                   call qtrap(ra,rb,s,debug)
                                   eff_bond = eff_bond + s
                                   eff_bond_total = eff_bond_total + s
                               else
                                   nerror = 2
                                   write(error(1),'(2a)') wfunc, ' weighting function is not supported in the code!'
                                   write(error(2),'(a)') 'Program aborted!'
                                   call erroroutput
                               end if
                           endif
                       endif

                       if ((dist_ri .gt. rsqsp) .and. (dist_rj .le. rsqsp)) then
                           ra = rj(:,m)
                           if (dist_rj - dist_r .gt. eps) then
                               eff_line = sqrt(rsqsp - dist_r) + &
                                          sign(sqrt(dist_rj - dist_r),1.0_dp-lambda)
                               rb = rj(:,m) + eff_line*(ri(:,m) - rj(:,m))/sqrt(dist_rij)
                               if (wfunc == 'constant') then
                                   eff_bond = eff_bond + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                                   eff_bond_total = eff_bond_total + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                               else if (wfunc == 'mollify') then
                                   call qtrap(ra,rb,s,debug)
                                   eff_bond = eff_bond + s
                                   eff_bond_total = eff_bond_total + s
                               else
                                   nerror = 2
                                   write(error(1),'(2a)') wfunc, ' weighting function is not supported in the code!'
                                   write(error(2),'(a)') 'Program aborted!'
                                   call erroroutput
                               end if
                           else
                               eff_line = sqrt(rsqsp - dist_r)
                               rb = rj(:,m) + eff_line*(ri(:,m) - rj(:,m))/sqrt(dist_rij)
                               if (wfunc == 'constant') then
                                   eff_bond = eff_bond + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                                   eff_bond_total = eff_bond_total + sqrt(ddot(dimn,ra-rb,1,ra-rb,1)) / hardy_volume
                               else if (wfunc == 'mollify') then
                                   call qtrap(ra,rb,s,debug)
                                   eff_bond = eff_bond + s
                                   eff_bond_total = eff_bond_total + s
                               else
                                   nerror = 2
                                   write(error(1),'(2a)') wfunc, ' weighting function is not supported in the code!'
                                   write(error(2),'(a)') 'Program aborted!'
                                   call erroroutput
                               end if
                           endif
                       endif
                   endif
               enddo

               if (eff_bond .gt. 0._dp) then
                   call tensor_prod(dimn,fij(:,p),rij,rcrossr)
                   tensor = tensor-eff_bond*rcrossr/sqrt(dist_rij)
               endif

           enddo
       endif

   enddo

   ! Collect the kinetic contribution 
   if (interior_cnt == 0) then
       avgvel = 0
   else
       avgvel = avgvel / interior_cnt
   endif

   do j = 1, interior_cnt
       i = interior(j)
       rel_vel = vel(:,i) - avgvel
       call tensor_prod(dimn,rel_vel,rel_vel,vcrossv)

       r = pos(:,i)-gpt(:,l)
       where (abs(r)/samplesize > 0.5_dp .and. pbc)
           r = r-sign(samplesize,r)
       end where
       dist = ddot(dimn,r,1,r,1)

       if (crossings) then
           if (wfunc == 'constant') then
               tensork = tensork + vcrossv * w_sphere_constant(sqrt(dist))
           else if (wfunc == 'mollify') then
               tensork = tensork + vcrossv * w_sphere_Nikhil(sqrt(dist))
           else
               nerror = 2
               write(error(1),'(2a)') wfunc, ' weighting function is not supported in the code!'
               write(error(2),'(a)') 'Program aborted!'
               call erroroutput
           end if
       else
           tensork = tensork + vcrossv / virial_volume
       endif
   enddo
   ! End of kinetic contribution
   end subroutine collect_hardy

    subroutine write_hardy_cauchy
    use mod_file, only : GetFreeUnit
    use mod_grid, only : ngrid, gpt
    use mod_output, only : string, output, heading
    use mod_rigidity, only : project

    implicit none
    integer :: unit, unit_kin, unit_i, unit_s, l

    heading = .true.
    write(string,'(a)') '**Writing Cauchy-Hardy stress tensor**'; call output
    write(string,'(a)') ''; call output

    unit = GetFreeUnit()
    open(unit,file='hardy_cauchy.str',status='replace',form='formatted')
    unit_kin = GetFreeUnit()
    open(unit_kin,file='hardy_kin_cauchy.str',status='replace',form='formatted')
    write(string,'(a)') 'Writing potential part of Cauchy-Hardy stress tensor to file hardy_cauchy.str'; call output
    write(string,'(a)') 'Writing kinetic part of Cauchy-Hardy stress tensor to file hardy_kin_cauchy.str'; call output
    if (project) then
        unit_i = GetFreeUnit()
        open(unit_i,file='hardy_i_cauchy.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected irrotational Cauchy-Hardy stress tensor to file hardy_i_cauchy.str'
        call output
        unit_s = GetFreeUnit()
        open(unit_s,file='hardy_s_cauchy.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected solenoidal Cauchy-Hardy stress tensor to file hardy_s_cauchy.str'
        call output
    end if

    do l = 1, ngrid
        write(unit,8010)       gpt(:,l), hardy(:,:,l)
        write(unit_kin,8010)   gpt(:,l), hardy_kin(:,:,l)
    enddo

    if (project) then
        do l = 1, ngrid
            write(unit_i,8010) gpt(:,l), hardy_i(:,:,l)
            write(unit_s,8010) gpt(:,l), hardy(:,:,l) - hardy_i(:,:,l)
        end do
    end if

    close(unit)
    close(unit_kin)
    if (project) then
        close(unit_i)
        close(unit_s)
    end if
    !8010 format(1x,12f25.15)
    8010 format(1x,12f25.16)

    end subroutine write_hardy_cauchy

    subroutine write_hardy_pk1
    use mod_file, only : GetFreeUnit
    use mod_grid, only : ngrid, gpt
    use mod_output, only : string, output, heading
    use mod_rigidity, only : project

    implicit none
    integer :: unit, unit_kin, unit_i, unit_s, l

    heading = .true.
    write(string,'(a)') '**Writing 1st pk-Hardy stress tensor**'; call output
    write(string,'(a)') ''; call output

    unit = GetFreeUnit()
    open(unit,file='hardy_pk1.str',status='replace',form='formatted')
    unit_kin = GetFreeUnit()
    open(unit_kin,file='hardy_kin_pk1.str',status='replace',form='formatted')
    write(string,'(a)') 'Writing potential part of 1st pk-Hardy stress tensor to file hardy_pk1.str'; call output
    write(string,'(a)') 'Writing kinetic part of 1st pk-Hardy stress tensor to file hardy_kin_pk1.str'; call output
    if (project) then
        unit_i = GetFreeUnit()
        open(unit_i,file='hardy_i_pk1.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected irrotational 1st pk-Hardy stress tensor to file hardy_i_pk1.str'
        call output
        unit_s = GetFreeUnit()
        open(unit_s,file='hardy_s_pk1.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected solenoidal 1st pk-Hardy stress tensor to file hardy_s_pk1.str'
        call output
    end if

    do l = 1, ngrid
        write(unit,8010)       gpt(:,l), hardy(:,:,l)
        write(unit_kin,8010)   gpt(:,l), hardy_kin(:,:,l)
    enddo

    if (project) then
        do l = 1, ngrid
            write(unit_i,8010) gpt(:,l), hardy_i(:,:,l)
            write(unit_s,8010) gpt(:,l), hardy(:,:,l) - hardy_i(:,:,l)
        end do
    end if

    close(unit)
    close(unit_kin)
    if (project) then
        close(unit_i)
        close(unit_s)
    end if
    !8010 format(1x,12f25.15)
    8010 format(1x,12f25.16)

    end subroutine write_hardy_pk1

    subroutine write_virial_cauchy
    use mod_file, only : GetFreeUnit
    use mod_grid, only : ngrid,gpt
    use mod_output, only : string, output, heading
    use mod_rigidity, only : project

    implicit none
    integer :: unit, unit_kin, unit_i, unit_s, l

    heading = .true.
    write(string,'(a)') '**Writing Cauchy-virial stress tensor**'; call output
    write(string,'(a)') ''; call output

    unit = GetFreeUnit()
    open(unit,file='virial_cauchy.str',status='replace',form='formatted')
    unit_kin = GetFreeUnit()
    open(unit_kin,file='virial_kin_cauchy.str',status='replace',form='formatted')
    write(string,'(a)') 'Writing potential part of Cauchy-virial stress tensor to file virial_cauchy.str'; call output
    write(string,'(a)') 'Writing kinetic part of Cauchy-virial stress tensor to file virial_kin_cauchy.str'
    call output
    if (project) then
        unit_i = GetFreeUnit()
        open(unit_i,file='virial_i_cauchy.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected irrotational Cauchy-virial stress tensor to file virial_i_cauchy.str'
        call output
        unit_s = GetFreeUnit()
        open(unit_s,file='virial_s_cauchy.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected solenoidal Cauchy-virial stress tensor to file virial_s_cauchy.str'
        call output
    end if

    do l = 1, ngrid
        write(unit,8010)       gpt(:,l), virial(:,:,l)
        write(unit_kin,8010)   gpt(:,l), virial_kin(:,:,l)
    enddo

    if (project) then
        do l = 1, ngrid
            write(unit_i,8010) gpt(:,l), virial_i(:,:,l)
            write(unit_s,8010) gpt(:,l), virial(:,:,l) - virial_i(:,:,l)
        end do
    end if

    close(unit)
    close(unit_kin)
    if (project) then
        close(unit_i)
        close(unit_s)
    end if
    !8010 format(1x,12f25.15)
    8010 format(1x,12f25.16)

    end subroutine write_virial_cauchy

    subroutine write_virial_pk1
    use mod_file, only : GetFreeUnit
    use mod_grid, only : ngrid,gpt
    use mod_output, only : string, output, heading
    use mod_rigidity, only : project

    implicit none
    integer :: unit, unit_kin, unit_i, unit_s, l

    heading = .true.
    write(string,'(a)') '**Writing 1st pk-virial stress tensor**'; call output
    write(string,'(a)') ''; call output

    unit = GetFreeUnit()
    open(unit,file='virial_pk1.str',status='replace',form='formatted')
    unit_kin = GetFreeUnit()
    open(unit_kin,file='virial_kin_pk1.str',status='replace',form='formatted')
    write(string,'(a)') 'Writing potential part of 1st pk-virial stress tensor to file virial_pk1.str'; call output
    write(string,'(a)') 'Writing kinetic part of 1st pk-virial stress tensor to file virial_kin_pk1.str'
    call output
    if (project) then
        unit_i = GetFreeUnit()
        open(unit_i,file='virial_i_pk1.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected irrotational 1st pk-virial stress tensor to file virial_i_pk1.str'
        call output
        unit_s = GetFreeUnit()
        open(unit_s,file='virial_s_pk1.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected solenoidal 1st pk-virial stress tensor to file virial_s_pk1.str'
        call output
    end if

    do l = 1, ngrid
        write(unit,8010)       gpt(:,l), virial(:,:,l)
        write(unit_kin,8010)   gpt(:,l), virial_kin(:,:,l)
    enddo

    if (project) then
        do l = 1, ngrid
            write(unit_i,8010) gpt(:,l), virial_i(:,:,l)
            write(unit_s,8010) gpt(:,l), virial(:,:,l) - virial_i(:,:,l)
        end do
    end if

    close(unit)
    close(unit_kin)
    if (project) then
        close(unit_i)
        close(unit_s)
    end if
    !8010 format(1x,12f25.15)
    8010 format(1x,12f25.16)

    end subroutine write_virial_pk1

    subroutine qtrap(ra,rb,s,toprint)
    use mod_box, only : dimn
    !use mod_output, only : string, output, erroroutput, nerror, error
    implicit none
    ! Transferred variables
    real(kind=dp), intent(in)  :: ra(dimn),rb(dimn)
    integer, intent(in)        :: toprint
    real(kind=dp), intent(out) :: s

    ! Local variables
    real(kind=dp), parameter :: eps=1.0e-3_dp
    integer, parameter:: jmax=30
    real(kind=dp) :: olds
    integer :: j

    olds = -1.0e30_dp
    s = 0
    do j = 1, jmax
        call trapzd(ra,rb,s,j,toprint)
! debug
!if (toprint == 1) then
!print *, 'times: ', j, ' s: ', s
!end if
! debug
        if ((j >= 5) .and. (abs(s-olds) .le. eps * abs(olds))) then
            !write(*,*) "Number of iterations: 2 power :",j
            return
        else if (j >= 5) then
            !print *, 'times: ', j, ' s: ', s
            !a = 1
        endif

        if (j == jmax) then
            if ((s < 1.0e-10_dp) .and. (olds < 1.0e-10_dp)) then
                s = 0.0_dp
                return
            else
                !nerror = 4
                !write(error(1),'(a)') 'NOT FULLY CONVERGED! SOMETHING WRONG! OR INCREASE jmax in mod_hardy.f90'
                !write(error(2),'(a,a)') 'If increasing jmax does not help, ', &
                !                         'you would probably need to modify the code to see what is going wrong!'
                !write(error(3),'(a,3F10.5,a,3F10.5)') 'ra = ', ra , ' rb = ', rb
                !write(error(3),'(a,3F10.5,a,3F10.5)') 's = ', s , ' olds = ', olds
                !call erroroutput

                write(*,'(a)') 'NOT FULLY CONVERGED! SOMETHING WRONG! OR INCREASE jmax in mod_hardy.f90'
                write(*,'(2a)') 'If increasing jmax does not help, ', &
                                'you would probably need to modify the code to see what is going wrong!'
                write(*,*) 'ra = ', ra , ' rb = ', rb
                write(*,*) 's = ', s , ' olds = ', olds
            endif
        endif

        olds=s
    enddo
    end subroutine qtrap

    subroutine trapzd(ra,rb,s,na,toprint)
    use mod_box, only : dimn
    use lsqrblasInterface, only : ddot
    implicit none
    ! Transferred variables
    real(kind=dp),intent(in)   :: ra(dimn),rb(dimn)
    integer, intent(in)        :: toprint
    integer, intent(in)        :: na
    real(kind=dp), intent(out) :: s
    ! Local variables
    real(kind=dp) :: x(dimn)
    real(kind=dp) :: da,db,dx,d,del,sum
    integer :: tnm,j
    integer, save :: it

    da = sqrt(ddot(dimn,ra,1,ra,1))
    db = sqrt(ddot(dimn,rb,1,rb,1))
    d = sqrt(ddot(dimn,ra-rb,1,ra-rb,1))

! debug
!if (toprint == 1) then
!print *, 'times: ', na
!print *, 'da: ', da
!print *, 'db: ', db
!print *, 'd: ', d
!end if
! debug

    if (d <= 1.0e-10_dp) then
        s = 0.0_dp
        return
    endif

    if (na .eq. 1) then
        s = 0.5_dp * d * (w_sphere_Nikhil(da) + w_sphere_Nikhil(db))

! debug
!if (toprint == 1) then
!print *, 'times: ', na
!print *, 'w_sphere(da): ', w_sphere_Nikhil(da)
!print *, 'w_sphere(db): ', w_sphere_Nikhil(db)
!print *, 'd: ', d
!print *, 's: ', s
!end if
! debug

        it=1
    else
        tnm = it
        del = d / tnm
        x = ra + 0.5_dp * del * (rb - ra) / d
        sum = 0.0_dp
        do j = 1, it
            dx = sqrt(ddot(dimn,x,1,x,1))
            sum = sum + w_sphere_Nikhil(dx)
            x = x + del * (rb - ra) / d
        enddo
        s = 0.5_dp * (s + d * sum / tnm)
        it = 2 * it
    endif
    return
    end subroutine trapzd

    real(kind=dp) function w_sphere_constant(r)
    use mod_output, only : nerror,error,erroroutput
    implicit none
    real(kind=dp), intent(in) :: r
    ! local variable

    if (r <= -epsilon(0.0_dp)) then
        nerror = 1
        error(1) = "Localization function's input (which is distance) should not be less than zero!"
        call erroroutput
    endif

    if (r .le. hardy_radius) then
        w_sphere_constant = 1.0_dp / hardy_volume
    else
        w_sphere_constant = 0.0_dp
    endif

    end function w_sphere_constant

    ! 
    ! This is a function in Nikhil's 2010 Paper
    ! Although there is some error in the paper, but the implementation here is correct.
    function w_sphere_Nikhil(r)
    use mod_global, only : PI
    use mod_output, only : nerror,error,erroroutput
    implicit none
    real(kind=dp),intent(in) :: r
    real(kind=dp) :: w_sphere_Nikhil
    real(kind=dp) :: factor,eps

    eps = 0.12_dp * hardy_radius
    !factor = 1._dp + (3._dp/PI**2 - 0.5_dp)*(eps/hardy_radius)**3 +    &
    !        (1.5_dp-6._dp/PI**2)*(eps/hardy_radius)**2 - 1.5_dp*(eps/hardy_radius)
    !factor = 1._dp/factor
    factor = 1.2011909586144_dp

    if (r <= -epsilon(0.0_dp)) then
        nerror = 1
        error(1) = "Localization function's input (which is distance) should not be less than zero!"
        call erroroutput
    endif

    if (r .le. (hardy_radius - eps)) then
        w_sphere_Nikhil = factor
    else if ((r .gt. (hardy_radius - eps)) .and. (r .le. hardy_radius)) then
        w_sphere_Nikhil = factor * 0.5_dp * (1.0_dp - cos((hardy_radius-r)*PI/eps))
    else
        w_sphere_Nikhil = 0.0_dp
    endif

    w_sphere_Nikhil = w_sphere_Nikhil / (4.0_dp*PI*(hardy_radius**3)/3.0_dp)

    end function w_sphere_Nikhil

end module mod_hardy
