!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**   MODULE mod_io: Handle input and output
!**
!**   type definitions:
!**   ----------------
!**   N/A
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   samplesize(:)
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   N/A
!**  
!**   logical Variable Definitions:
!**   ----------------------------
!**   N/A
!**
!**   Contains Routines:
!**   ------------------
!**   read_input : reads in parameters given by the user
!****************************************************************************
module mod_io
implicit none
save
private
public read_input
contains
    !-------------------------------------------------------------------------------
    ! read_input  :  reads the input file
    !
    subroutine read_input
    ! allocate_box reads box in config file
    ! allocate_species reads species in species file
    use mod_box, only : allocate_box, allocate_species
    ! read_config reads config in config file
    use mod_particles, only : read_config
    use mod_output, only : erroroutput,nerror,error
    use mod_global, only : cmdunit
    use mod_qclib, only : next,downcase
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=512) :: filename_species,filename_config,input
    character(len=4) :: cc
    integer :: lower,upper,ldif,ierror
    logical :: spec_called,conf_called

    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    spec_called=.false.
    conf_called=.false.
    filename_species = 'species.data'
    filename_config = 'config.data'
    do
        read(cmdunit,'(a512)',iostat=ierror) input
        if (ierror .ne. 0) then
            nerror = 2
            write(error(1),'(a)') 'Something wrong with the command input.'
            write(error(2),'(a)') 'Macro end must be called to continue the simulation.'
            call erroroutput
        end if
        lower = 0
        upper = next(lower,input)
        cc = input(lower+1:upper-1)

        select case(trim(downcase(cc)))
        case('conf')
            if (conf_called) then
                nerror = 2
                write(error(1),'(a)') 'Configuration file has already been given!'
                write(error(2),'(a)') 'config could not be called twice!'
                call erroroutput
            end if

            lower=upper
            upper=next(lower,input)
            ldif = upper - lower - 1
            if (ldif /= 0) then
                filename_config = trim(input(lower+1:upper-1)) // '.data'
            end if
            if (spec_called) then
                call allocate_box(filename_config)
                ! set a flag that dimensions and pbc are read in
                call read_config(filename_config)
            else
                nerror = 1
                write(error(1),'(a)') "Macro spec must be called before calling conf."
                call erroroutput
            end if
            conf_called=.true.
        case('spec')
            if (spec_called) then
                nerror = 2
                write(error(1),'(a)') 'Species file has already been given!'
                write(error(2),'(a)') 'species could not be called twice!'
                call erroroutput
            end if
            lower=upper
            upper=next(lower,input)
            ldif = upper - lower - 1
            if (ldif /= 0) then
                filename_species = trim(input(lower+1:upper-1)) // '.data'
            end if
            call allocate_species(filename_species)
            spec_called=.true.
        case('end')
            if (spec_called .and. conf_called) then
                exit
            else
                if (spec_called .and. (.not. conf_called)) then
                    nerror = 1
                    write(error(1),'(a)') "Macro conf must be called in stage read."
                    call erroroutput
                else
                    nerror = 1
                    write(error(1),'(a)') "Macro spec and conf must be called in stage read."
                    call erroroutput
                end if
            end if
        case default
            nerror = 1
            error(1) = 'macro "' // cc // '" not recognized in stage read**'
            call erroroutput
        end select
    end do

    end subroutine read_input
end module mod_io
