!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!-------------------------------------------------------------------------------
! MODULE mod_cayley :
!
! type definitions:
! ----------------
! N/A
!
! Real Variable Definitions:
! -------------------------------------
! 
!
! Integer Variable Definitions:
! ----------------------------
!
! logical Variable Definitions: N/A
! ----------------------------
!
! Contains Routines:
! ------------------
!
#include "KIM_API_status.h"
#define THIS_FILE_NAME __FILE__
#define TRUEFALSE(TRUTH) merge(1,0,(TRUTH))
! means if TRUTH = TRUE -> Result = 1
!       if TRUTH = FALSE -> Result = 0 
! This is not used anywhere
module mod_kim

use, intrinsic :: iso_c_binding
use KIM_API_F03
use mod_types
use mod_global, only : dp
use mod_output, only : string, output
implicit none
save
public
type(c_ptr) :: pkim
integer(c_int) :: ier,idum

contains
    subroutine setup_KIM_API_object(testkimfile, modelname,  &
                                    dimn,nat,nsp,pos,element,particle_energy,particleTypes,energy,fi,rcutoff,samplesize,nbc)
    !-------------------------------------------------------------------------------
    !
    ! free_KIM_API_object : Deallocate storage and destroy KIM API object
    !
    !-------------------------------------------------------------------------------
    use mod_output, only : erroroutput,nerror,error
    implicit none

    !-- Transferred variables
    character(len=KIM_KEY_STRING_LENGTH), intent(in)    :: testkimfile
    character(len=KIM_KEY_STRING_LENGTH), intent(in)    :: modelname
    character(len=KIM_KEY_STRING_LENGTH)                :: NBC_Method
    integer                                             :: nbc
    integer,                              intent(in)    :: dimn
    integer,        target,               intent(in)    :: nat
    integer,        target,               intent(in)    :: nsp
    real(kind=dp),  target,               intent(in)    :: pos(dimn,nat)
    character(len=3),                     intent(in)    :: element(nat)
    real(kind=dp),  target,               intent(in)    :: particle_energy(nat),energy,fi(dimn,nat)
    integer(c_int), target,               intent(inout) :: particleTypes(nat)

    integer(c_int), parameter                           :: SizeOne = 1
    real(kind=dp),  target,               intent(out)   :: rcutoff
    real(kind=dp),  target,               intent(in)    :: samplesize(dimn)

    !-- Local variables
    real(c_double), pointer                             :: cutoff
    type(c_ptr)                                         :: pcutoff
    integer                                             :: i

    particleTypes=1

    call Write_KIM_descriptor_file(testkimfile)

    ! Initialize KIM API object
    !
    ier = kim_api_file_init(pkim, testkimfile, modelname)
    if (ier .lt. KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, THIS_FILE_NAME, "kim_api_file_init", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_FILE_INIT ERROR!'
        call erroroutput
    endif

    call kim_api_setm_data( pkim,ier,                              &
        "numberOfParticles",           SizeOne, c_loc(nat),      1, &
        "numberContributingParticles", SizeOne, c_loc(nat),      1, &
        "coordinates",                 dimn*nat,c_loc(pos),      1, &
        "particleSpecies",             nat,     c_loc(particleTypes),1, &
        "forces",                      dimn*nat,c_loc(fi),       1, &
        "particleEnergy",              nat    , c_loc(particle_energy),   1, &
        "energy",                      SizeOne, c_loc(energy),   1, &
        !"boxSideLengths",              dimn,    c_loc(samplesize),TRUEFALSE((nbc .eq. 0).or.(nbc.eq.1)), &
        "boxSideLengths",              dimn,    c_loc(samplesize),1,&
        "cutoff",                      SizeOne, c_loc(rcutoff),1,&
        "numberOfSpecies",             SizeOne, c_loc(nsp),      1)

    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, __FILE__, "kim_api_setm_data", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_SETM_DATA ERROR!'
        call erroroutput
    endif

    ! call model's init routine
    !
    ier = kim_api_model_init(pkim)
    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, __FILE__, "kim_api_model_init", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_MODEL_INIT ERROR!'
        call erroroutput
    endif

    ! determine which NBC scenerio to use
    ier = kim_api_get_nbc_method(pkim, NBC_Method)
    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, THIS_FILE_NAME, &
                                 "kim_api_get_nbc_method", ier)
        nerror = 1 
        write(error(1),'(a)') 'KIM_API_GET_NBC_METHOD ERROR!'
        call erroroutput
    endif

    if (index(NBC_Method,"MI_OPBC_H").eq.1) then
        write(string,'(a)') 'Your model is using MI_OPBC_H'; call output
        nbc = 0
    elseif (index(NBC_Method,"MI_OPBC_F").eq.1) then
        write(string,'(a)') 'Your model is using MI_OPBC_F'; call output
        nbc = 1
    elseif (index(NBC_Method,"NEIGH_PURE_H").eq.1) then
        write(string,'(a)') 'Your model is using NEIGH_PURE_H'; call output
        nbc = 2
    elseif (index(NBC_Method,"NEIGH_PURE_F").eq.1) then
        write(string,'(a)') 'Your model is using NEIGH_PURE_F'; call output
        nbc = 3
    else
        ier = KIM_STATUS_FAIL
        idum = kim_api_report_error(__LINE__, THIS_FILE_NAME, &
                 "MDStressLab only supports MI_OPBC_H, MI_OPBC_F, NEIGH_PURE_H and NEIGH_PURE_F", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_GET_NBC_METHOD ERROR!'
        call erroroutput
    endif

    do i = 1, nat
        particleTypes(i) = kim_api_get_species_code(pkim,element(i),ier)
    enddo

    return
    end subroutine setup_KIM_API_object

    subroutine setup_kim_access(pcalc_fij,pget_neigh,pneighObjFull)
    use mod_output, only : erroroutput,nerror,error
    !-------------------------------------------------------------------------------
    !
    ! setup_neighborlist_no_Rij_KIM_access :
    !
    !    Store necessary pointers in KIM API object to access the neighbor list
    !    data and methods.
    !
    !-------------------------------------------------------------------------------
    implicit none

    !-- Transferred variables
    type(c_funptr), intent(in) :: pcalc_fij,pget_neigh
    type(c_ptr), intent(in)    :: pneighObjFull

    !-- Local variables
    integer(c_int), parameter  :: SizeOne = 1

    ! store pointers to neighbor list object and access function
    !
    call kim_api_setm_method(pkim, ier,                &
       "get_neigh",     SizeOne, pget_neigh,    1,     &
       "process_dEdr",  SizeOne, pcalc_fij,     1)

    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, __FILE__, "kim_api_setm_data", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_SETM_METHOD ERROR!'
        call erroroutput
    endif

    ier = kim_api_set_data(pkim, "neighObject", SizeOne, pneighObjFull)
    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, THIS_FILE_NAME, &
                                    "kim_api_set_data", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_SET_DATA ERROR!'
        call erroroutput
    endif

    return
    end subroutine setup_kim_access

    subroutine model_compute
    use mod_output, only : erroroutput,nerror,error
    implicit none

    ier = KIM_API_model_compute(pkim)
    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, __FILE__, "kim_api_model_compute", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_MODEL_COMPUTE ERROR!'
        call erroroutput
    endif
    end subroutine model_compute

    subroutine free_KIM_API_object
    use mod_output, only : erroroutput,nerror,error
    !-------------------------------------------------------------------------------
    !
    ! free_KIM_API_object : Deallocate storage and destroy KIM API object
    !
    !-------------------------------------------------------------------------------
    implicit none

    ! call the model destroy function
    !
    ier = kim_api_model_destroy(pkim)
    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, __FILE__, "kim_api_model_destroy", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_MODEL_DESTROY ERROR!'
        call erroroutput
    endif

    ! free all KIM API object storage
    !
    call kim_api_free(pkim, ier)
    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, __FILE__, "kim_api_free", ier)
        nerror = 1
        write(error(1),'(a)') 'KIM_API_FREE ERROR!'
        call erroroutput
    endif

    return

    end subroutine free_KIM_API_object

!-------------------------------------------------------------------------------
!
!  Write KIM descriptor file for MDStressLab
!
!-------------------------------------------------------------------------------
subroutine Write_KIM_descriptor_file(kim_descriptor_file)
use mod_box,   only : spec,nsp
use mod_file, only : GetFreeUnit
implicit none

!-- Transferred variables
character(len=KIM_KEY_STRING_LENGTH), intent(in) :: kim_descriptor_file

!-- Local variables
integer :: i, dunit
character(len=10000) :: kim_descriptor_file_string
character(len=103)   :: divider
character(len=1)     :: cr
character(len=52)    :: type_line

! Define frequently used variables
!
cr = char(10)
divider = '#######################################################################################################'

! Write Minimol descriptor file into string kim_descriptor
!
kim_descriptor_file_string = &
  divider                                                                            // cr // &
  '#'                                                                                // cr // &
  '# Copyright 2017 Ellad B. Tadmor, Ryan S. Elliott, and James P. Sethna'           // cr // &
  '# All rights reserved.'                                                           // cr // &
  '#'                                                                                // cr // &
  '# Author: Automatically generated by MDStressLab'                                 // cr // &
  '#'                                                                                // cr // &
  '#'                                                                                // cr // &
  '# See KIM_API/standard.kim for documentation about this file'                     // cr // &
  '#'                                                                                // cr // &
  divider                                                                            // cr // &
                                                                                        cr // &
                                                                                        cr // &
  'KIM_API_Version  := 1.7.3'                                                         // cr // &
                                                                                        cr // &
  'Unit_length      := A'                                                            // cr // &
  'Unit_energy      := eV'                                                           // cr // &
  'Unit_charge      := e'                                                            // cr // &
  'Unit_temperature := K'                                                            // cr // &
  'Unit_time        := ps'                                                           // cr // &
                                                                                        cr // &
                                                                                        cr // &
  divider                                                                            // cr // &
  'PARTICLE_SPECIES:'                                                  // cr // &
  '# Symbol/name           Type                    code'                             // cr

do i = 1,nsp
    write(type_line,'(a3,21x,''spec'',20x,i4)') spec(i),i
    kim_descriptor_file_string = trim(kim_descriptor_file_string) // type_line // cr
enddo

kim_descriptor_file_string = trim(kim_descriptor_file_string) // &
                                                                                        cr // &
                                                                                        cr // &
  divider                                                                            // cr // &
  'CONVENTIONS:'                                                                     // cr // &
  '# Name                  Type'                                                     // cr // &
                                                                                        cr // &
  'OneBasedLists           flag'                                                     // cr // &
                                                                                        cr // &
  'Neigh_LocaAccess        flag'                                                     // cr // &
                                                                                        cr // &
  'MI_OPBC_H               flag'                                                     // cr // &
                                                                                        cr // &
  'MI_OPBC_F               flag'                                                     // cr // &
                                                                                        cr // &
  'NEIGH_PURE_H            flag'                                                     // cr // &
                                                                                        cr // &
  'NEIGH_PURE_F            flag'                                                     // cr // &
                                                                                        cr // &
                                                                                        cr // &
  divider                                                                            // cr // &
  'MODEL_INPUT:'                                                                     // cr // &
  '# Name                        Type         Unit       Shape              requirements' // cr // &
  'numberOfParticles             integer      none       []'                         // cr // &
                                                                                        cr // &
  'numberContributingParticles   integer      none       []'                         // cr // &
                                                                                        cr // &
  'numberOfSpecies               integer      none       []'                         // cr // &
                                                                                        cr // &
  'particleSpecies               integer      none       [numberOfParticles]'        // cr // &
                                                                                        cr // &
  'coordinates                   double       length     [numberOfParticles,3]'      // cr // &
                                                                                        cr // &
  'process_dEdr                  method       none       []'                         // cr // &
                                                                                        cr // &
  'get_neigh                     method       none       []'                         // cr // &
                                                                                        cr // &
  'neighObject                   pointer      none       []'                         // cr // &
                                                                                        cr // &
  'boxSideLengths                double       length     [3]               optional' // cr // &
                                                                                        cr // &
                                                                                        cr // &
  divider                                                                            // cr // &
  'MODEL_OUTPUT:'                                                                    // cr // &
  '# Name                        Type         Unit       Shape              requirements' // cr // &
                                                                                        cr // &
  'destroy                       method       none       []'                         // cr // &
                                                                                        cr // &
  'compute                       method       none       []'                         // cr // &
                                                                                        cr // &
  'reinit                        method       none       []'                         // cr // &
                                                                                        cr // &
  'cutoff                        double       length     []'                         // cr // &
                                                                                        cr // &
  'energy                        double       energy     []'                         // cr // &
                                                                                        cr // &
  'particleEnergy                double       energy     [numberOfParticles]'        // cr // &
                                                                                        cr // &
  'forces                        double       force      [numberOfParticles,3]'      // cr // &
                                                                                        cr // &
  divider                                                                            // cr 
  !// char(0)

! Write the string kim_descriptor_file_string to descriptor.kim 
dunit=GetFreeUnit()
open(dunit,file=trim(kim_descriptor_file),action='write')
write(dunit,'(a)') trim(kim_descriptor_file_string)
close(dunit)

return

end subroutine Write_KIM_descriptor_file
end module mod_kim
