!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2017, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Min Shi
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!-------------------------------------------------------------------------------
! MODULE :
!
! type definitions:
! ----------------
! N/A
!
! Real Variable Definitions:
! -------------------------------------
!
! Integer Variable Definitions:
! ----------------------------
!
! logical Variable Definitions:
! ----------------------------
!
! Contains Routines:
! ------------------
!
!
!
!
! SUBROUTINE :
!
! Passed Parameters :-
! -----------------
!
! Non-Obvious Parameter :-
! ---------------------
!
module mod_tsai
use mod_global, only : dp
implicit none
save
private
public tsai,              &
       tsai_i,            &
       collect_tsai,      &
       write_tsai_cauchy, &
       write_tsai_pk1,    &
       tsai_length,       &
       tsai_area,         &
       tsai_radius,       &
       avgdomain_tsai

real(kind=dp), dimension(:,:,:), allocatable :: tsai, tsai_i
real(kind=dp)                                :: tsai_area, tsai_length

character(len=512)              :: avgdomain_tsai = 'sphere'

! For spherical average domain
real(kind=dp)                   :: tsai_radius = 1.0_dp

contains                                              

    subroutine collect_tsai(l,cutoff,fij,list,marker1,marker2,tensor,avgdomain,wfunc)
    use mod_box, only : dimn,samplesize,pbc!,nat
    use mod_grid, only : gpt
    use mod_particles, only : pos
    use mod_neigh, only : full_grid_verlet_list
    use mod_qclib, only : tensor_prod
    use lsqrblasInterface, only : ddot
    implicit none
    ! Passed variables
    integer,            intent(in)  :: l
    integer,            intent(in)  :: marker1(:),marker2(:),list(:)
    real(kind=dp),      intent(in)  :: cutoff,fij(:,:)
    character(len=512), intent(in)  :: avgdomain,wfunc
    real(kind=dp),      intent(out) :: tensor(:,:)

    integer :: i,j,k,p,m,n
    real(kind=dp) :: dist,eps = 1.0e-5_dp,dist_ri,dist_rj,rsumsq
    real(kind=dp), dimension(dimn)   :: r,p1,p2,rij,rp1,rp2
    real(kind=dp), dimension(dimn,2) :: ri,rj

    tensor = 0.0_dp
    rsumsq = (tsai_radius + cutoff) ** 2
    eps = 1.0e-5_dp

    !do i = 1, nat ! n=1,full_grid_verlet_list(l)%nneigh
        !i = full_grid_verlet_list(l)%neigh(n)
    do n = 1, full_grid_verlet_list(l)%nneigh
        i = full_grid_verlet_list(l)%neigh(n)
        r = pos(:,i)-gpt(:,l)
        where (abs(r)/samplesize > 0.5_dp .and. pbc)
            r = r-sign(samplesize,r)
        end where
        dist = ddot(dimn,r,1,r,1)

        if (dist .le. rsumsq) then
            do p=marker1(i),marker2(i)
            !-------------------------------------------------------------------------------
            !
            !                 ==============================
            !                ||                            ||
            !                ||  _______                   ||
            !                || / .p1   \                  ||  .rp1
            !                ||/ /       \                 || /
            !                ||\/  Avg.  /                 ||/
            !                ||/\ domain/                  |/
            !                |/  \     /                   /|
            !                /|   \___/                   /||
            !               /||                          / ||
            !           rp2. ||                       p2.  ||
            !                ||                            ||
            !                 ==============================
            !
            !
                j = list(p)
                p1 = pos(:,i)
                p2 = pos(:,j)
                rij = p1-p2
                where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                    rij = rij-sign(samplesize,rij)
                end where
                rp1 = p2 + rij
                rp2 = p1 - rij
                if (maxval(abs(rp1-p1)) .lt. eps) then
                    k=1
                else
                    k=2
                endif

                ri(:,1) = p1  - gpt(:,l)
                rj(:,1) = rp2 - gpt(:,l)
                ri(:,2) = rp1 - gpt(:,l)
                rj(:,2) = p2  - gpt(:,l)
                                        
                do m = 1,k
                    dist_ri = ddot(dimn,ri(:,m),1,ri(:,m),1)
                    dist_rj = ddot(dimn,rj(:,m),1,rj(:,m),1)
                    if (dist_ri .gt. rsumsq .or. dist_rj .gt. rsumsq) then
                        cycle
                    endif

                    if (line_crosses_surface(ri(:,m),rj(:,m),1)) then
                       if (ri(1,m) < 0.0_dp) then
                         tensor(:,1) = tensor(:,1) + fij(:,p)
                       else
                         tensor(:,1) = tensor(:,1) - fij(:,p)
                       endif
                    endif
                    if (line_crosses_surface(ri(:,m),rj(:,m),2)) then
                       if (ri(2,m) < 0.0_dp) then
                          tensor(:,2) = tensor(:,2) + fij(:,p)
                       else
                          tensor(:,2) = tensor(:,2) - fij(:,p)
                      endif
                    endif
                    if (line_crosses_surface(ri(:,m),rj(:,m),3)) then
                       if (ri(3,m) < 0.0_dp) then
                          tensor(:,3) = tensor(:,3) + fij(:,p)
                       else
                          tensor(:,3) = tensor(:,3) - fij(:,p)
                       endif
                    endif
                enddo
            enddo
        endif
    enddo

    tensor = tensor / tsai_area

    end subroutine collect_tsai

    logical function line_crosses_surface(ri,rj,surf)
    !
    ! returns true if the line defined by pos1 and pos2
    ! crosses the surface defined in module Tsai
    !
    implicit none

    !Passed variables
    real(kind=dp), intent(in) :: ri(:),rj(:)
    integer,       intent(in) :: surf

    !Local variables
    real(kind=dp), parameter  :: tol=1.0e-7_dp
    real(kind=dp)             :: posint_x,posint_y,posint_z,slope,rij(3)

    Line_Crosses_Surface = .false.
    rij = ri - rj
    ! surf = 1 -> Surface normal to x-axis
    if(surf .eq. 1) then
        if (rj(1) * ri(1) < 0.0_dp) then
            slope = rij(2)/rij(1)
            posint_y = ri(2) - slope*ri(1)
            slope = rij(3)/rij(1)
            posint_z = ri(3) - slope*ri(1)
            if (posint_y > -0.5_dp*tsai_length-tol .and. &
                posint_y <  0.5_dp*tsai_length+tol .and. &
                posint_z > -0.5_dp*tsai_length-tol .and. &
                posint_z <  0.5_dp*tsai_length+tol) then
                Line_Crosses_Surface = .true.
            endif
        endif
    ! surf = 2 -> Surface normal to y-axis
    elseif(surf .eq. 2) then
        if (rj(2) * ri(2) < 0.0_dp) then
             slope = rij(1)/rij(2)
             posint_x = ri(1) - slope*ri(2)
             slope = rij(3)/rij(2)
             posint_z = ri(3) - slope*ri(2)
             if (posint_x > -0.5_dp*tsai_length-tol .and. &
                 posint_x <  0.5_dp*tsai_length+tol .and. &
                 posint_z > -0.5_dp*tsai_length-tol .and. &
                 posint_z <  0.5_dp*tsai_length+tol) then
                 Line_Crosses_Surface = .true.
             endif
        endif
    ! surf = 3 -> Surface normal to y-axis
    elseif(surf .eq. 3) then
        if (rj(3)*ri(3) < 0.0_dp) then
            slope = rij(1)/rij(3)
            posint_x = ri(1) - slope*ri(3)
            slope = rij(2)/rij(3)
            posint_y = ri(2) - slope*ri(3)
            if (posint_x > -0.5_dp*tsai_length-tol .and. &
                posint_x <  0.5_dp*tsai_length+tol .and. &
                posint_y > -0.5_dp*tsai_length-tol .and. &
                posint_y <  0.5_dp*tsai_length+tol) then
                Line_Crosses_Surface = .true.
            endif
        endif
    endif

    end function Line_Crosses_Surface

    subroutine write_tsai_cauchy
    use mod_file, only : GetFreeUnit
    use mod_grid, only : ngrid, gpt
    use mod_output, only : string, output, heading
    use mod_rigidity, only : project
    implicit none
    integer :: unit, unit_i, unit_s, l

    heading = .true.
    write(string,'(a)') '**Writing Cauchy-Tsai stress tensor**'; call output
    write(string,'(a)') ''; call output

    unit = GetFreeUnit()
    open(unit,file='tsai_cauchy.str',status='replace',form='formatted')
    write(string,'(a)') 'Writing potential part of Cauchy-Tsai stress tensor to file tsai_cauchy.str.'; call output
    if (project) then
        unit_i = GetFreeUnit()
        open(unit_i,file='tsai_i_cauchy.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected irrotational Cauchy-Tsai stress tensor to file tsai_i_cauchy.str.'
        call output

        unit_s = GetFreeUnit()
        open(unit_s,file='tsai_s_cauchy.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected solenoidal Cauchy-Tsai stress tensor to file tsai_s_cauchy.str.'
        call output
    end if

    do l = 1, ngrid
        write(unit,8010)         gpt(:,l), tsai(:,:,l)
    enddo

    if (project) then
        do l = 1, ngrid
            write(unit_i,8010)   gpt(:,l), tsai_i(:,:,l)
            write(unit_s,8010)   gpt(:,l), tsai(:,:,l) - tsai_i(:,:,l)
        end do
    end if

    close(unit)
    if (project) then
        close(unit_i)
        close(unit_s)
    end if

    !8010 format(1x,12f25.15)
    8010 format(1x,12f25.16)

    end subroutine write_tsai_cauchy

    subroutine write_tsai_pk1
    use mod_file, only : GetFreeUnit
    use mod_grid, only : ngrid, gpt
    use mod_output, only : string, output, heading
    use mod_rigidity, only : project
    implicit none
    integer :: unit, unit_i, unit_s, l

    heading = .true.
    write(string,'(a)') '**Writing 1st pk-Tsai stress tensor**'; call output
    write(string,'(a)') ''; call output

    unit = GetFreeUnit()
    open(unit,file='tsai_pk1.str',status='replace',form='formatted')
    write(string,'(a)') 'Writing potential part of 1st pk-Tsai stress tensor to file tsai_pk1.str.'; call output
    if (project) then
        unit_i = GetFreeUnit()
        open(unit_i,file='tsai_i_pk1.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected irrotational 1st pk-Tsai stress tensor to file tsai_i_pk1.str.'
        call output

        unit_s = GetFreeUnit()
        open(unit_s,file='tsai_s_pk1.str',status='replace',form='formatted')
        write(string,'(a)') 'Writing projected solenoidal 1st pk-Tsai stress tensor to file tsai_s_pk1.str.'
        call output
    end if

    do l = 1, ngrid
        write(unit,8010)         gpt(:,l), tsai(:,:,l)
    enddo

    if (project) then
        do l = 1, ngrid
            write(unit_i,8010)   gpt(:,l), tsai_i(:,:,l)
            write(unit_s,8010)   gpt(:,l), tsai(:,:,l) - tsai_i(:,:,l)
        end do
    end if

    close(unit)
    if (project) then
        close(unit_i)
        close(unit_s)
    end if

    !8010 format(1x,12f25.15)
    8010 format(1x,12f25.16)

    end subroutine write_tsai_pk1

end module mod_tsai

